<?php

namespace Intergo\RedisProxy\Proxies;

use Illuminate\Cache\CacheManager;
use Illuminate\Contracts\Cache\Repository;
use Illuminate\Redis\Connections\Connection;
use Illuminate\Redis\RedisManager;
use Illuminate\Support\Facades\Log;
use Throwable;

abstract class OperationsProxy implements ProxyOperationsInterface
{

    const DEFAULT_CONFIG_KEY = 'default';
    const READONLY_CONFIG_KEY = 'read';

    /**
     * Connection retries for failover to default connection
     *
     * @var int
     */
    protected $connectionsRetries = 2;

    /**
     * Proxy call on read or default connection/store based on called method
     *
     * @param string $method
     * @param array $parameters
     * @return mixed
     */
    abstract protected function proxy(string $method, array $parameters);

    /**
     * @param CacheManager|RedisManager $manager
     * @param string $connection
     * @return Repository|Connection|null
     */
    protected function readConnect($manager, string $connection)
    {
        $readConnection = null;
        $tries = 0;
        do {
            try {
                //cache & redis managers implement different contracts
                if ($manager instanceof CacheManager) {
                    $readConnection = $manager->store($connection);
                } elseif ($manager instanceof RedisManager) {
                    $readConnection = $manager->connection($connection);
                } else {
                    return null;
                }
            } catch (Throwable $exception) {
                $tries++;
                //sleep here to give some time for redis to establish back the connection
                sleep(1);
                if ($tries == $this->connectionsRetries) {
                    //notify for failure and continue using master connection
                    Log::error("Redis Read ({$connection}) failed to connect: " . $exception->getMessage());
                }
            }
        } while ($tries < $this->connectionsRetries && is_null($readConnection));

        return $readConnection;
    }

    /**
     * Retry executing call on read only cache store/redis connection
     * @param Repository|Connection $readProxy
     * @param string $method
     * @param array $parameters
     * @return array
     */
    protected function failover($readProxy, string $method, array $parameters) {
        $tries = 0;
        $connected = false;
        $result = null;

        if (is_null($readProxy)) {
            return [$connected, $result];
        }

        do {
            try {
                $result = $readProxy->{$method}(...$parameters);
                $connected = true;
            } catch (Throwable $exception) {
                $tries++;
            }
        } while ($tries < $this->connectionsRetries && !$connected);

        return [$connected, $result];

    }
}
